#!/usr/bin/env perl

use XyceRegression::Tools;
use Scalar::Util qw(looks_like_number);

$Tools = XyceRegression::Tools->new();
#$Tools->setDebug(1);
#$Tools->setVerbose(1);

if (defined($verbose)) { $Tools->setVerbose(1); }

# The input arguments to this script are: 
# $1 = location of Xyce binary
# $2 = location of xyce_verify.pl script
# $3 = location of compare script 
# $4 = location of circuit file to test
# $5 = location of gold standard prn file

# If Xyce does not produce a prn file, then we return exit code 10.
# If Xyce succeeds, but the test fails, then we return exit code 2.
# If the shell script fails for some reason, then we return exit code 1.
# Otherwise we return the exit code of compare or xyce_verify.pl

# Since the shell script runs Xyce and the comparison program, it is
# responsible for capturing any error output from Xyce and the STDOUT & STDERR
# from the comparison program.  The outside script, run_xyce_regression,
# expects the STDERR output from Xyce to go into $CIRFILE.err, the STDOUT
# output from comparison to go into $CIRFILE.prn.out and the STDERR output from
# comparison to go into $CIRFILE.prn.err.  

use Getopt::Long;
use File::Basename;

&GetOptions( "verbose!" => \$verbose );
$XYCE=$ARGV[0];
$XYCE_VERIFY=$ARGV[1];
#$XYCE_COMPARE=$ARGV[2];
$CIRFILE=$ARGV[3];
$GOLDPRN=$ARGV[4];

# Comparison tolerances
my $absTol=1e-5;
my $relTol=1e-3;
my $zeroTol=1e-10;

# create the output files
my $CIRFILE="bug_646.cir";
$retval=$Tools->wrapXyce($XYCE,$CIRFILE);

my $foundDataStart=0;
# check each output file
foreach $idx (1 .. 10)
{
  # file names
  my $testFileName="$CIRFILE$idx.csd";
  my $goldFileName="$CIRFILE$idx.GSfile";

  if ( ! -e $testFileName)
  {
    print "Failed to produce output file \n";
    print "Exit code = 14\n";
    exit 14;
  }

  # open files and check that the files have the same number of lines
  if (not -s "$goldFileName" ) 
  { 
    print STDERR "Missing Gold Standard file: $goldFileName\nExit code = 2\n"; 
    $exitCode=2; 
    print "test Failed!\n";
    print "Exit code = $exitCode\n";
    exit $exitCode;
  }
  if (not -s "$testFileName" ) 
  { 
    print STDERR "Missing Test file: $testFileName\nExit code = 14\n"; 
    $exitCode=14; 
    print "test Failed!\n";
    print "Exit code = $exitCode\n";
    exit $exitCode;
  }

  open(TESTFILE,"$testFileName");
  open(GSFILE,"$goldFileName");

  # first test if the files have the same number of lines
  for ($testFileCount=0; <TESTFILE>; $testFileCount++) { }
  for ($gsCount=0; <GSFILE>; $gsCount++) { }

  close(GSFILE);
  close(TESTFILE);

  $exitCode=0;
  if ($testFileCount != $gsCount) 
  {
    print STDERR "file $testFileName doesn't match the Gold Standard\n";
    print STDERR "$testFileName line count= $testFileCount\n";
    print STDERR "Gold Standard line count= $gsCount\n";
    $exitCode=2;
  }
  else
  { 
    # If the line counts match, then compare in detail.
    # Re-open the files and compare them line by line.
    open(TESTFILE,"$testFileName");
    open(GSFILE,"$goldFileName");
    $lineCount=0;
    while( ($lineGS=<GSFILE>) && ($lineTestFile=<TESTFILE>) )
    {
      $lineCount++;
      # process a line into text and values.
      chop $lineGS;
      # Remove leading spaces on lineGS, otherwise the spaces become 
      # element 0 of "@gsData" instead of the first column of data.
      $lineGS =~ s/^\s*//;
      @gsData = (split(/[\s,]+/, $lineGS));

      # process a line into text and values.
      chop $lineTestFile;
      # Remove leading spaces on line, otherwise the spaces become 
      # element 0 of "testFileData" instead of the first column of data.
      $lineTestFile =~ s/^\s*//;
      @testFileData = (split(/[\s,]+/, $lineTestFile));

      if ($lineTestFile =~ /#C/)
      {
        $foundDataStart=1;
      }
           
      if ($#gsData != $#testFileData)
      {
        print STDERR "File $testFileName and Gold Standard have a different # of elements on line $lineCount:\n";
        print STDERR "File $testFileName had $#testFileData elements\n";
        print STDERR "Gold Standard had $#gsData elements\n";
        $exitCode=2;
      }
      # skip version and date/time lines, since they may change in each
      # test run.  Also need to handle XBEGIN line separately
      elsif ( ($lineTestFile =~ /SOURCE/) || ($lineTestFile =~ /TIME/) )
      {
        #no op
      }
      else
      {
        # the two files have the same number of items on a line.  
        # compare individual values as scalars or strings  
        for( $i=0; $i<=$#testFileData; $i++ )
        {
          if ( ($testFileData[$i] =~ /XBEGIN/) || ($testFileData[$i] =~ /XEND/) )
          {
            # XBEGIN line has numbers formatted as XBEGIN='0.00000000e+00' XEND='1.00000000e-02'
            # Need to extract the numbers.
            @testData = split(/[\']/,$testFileData[$i]);
            @goldData = split(/[\']/,$gsData[$i]);
            if ($#gsData != $#testFileData)
            {
              print STDERR "File $testFileName and Gold Standard have mismatched XBEGIN line\n";
              last;
              $exitCode=2;
            }
            if( ($testData[0] eq $goldData[0]) ) 
            {
              # two string fields match for XBEGIN and XEND
            }
            else
            {
              print STDERR "String elements failed compare on line $lineCount:\n";
              print STDERR "File $testFileName produced: \"$testFileData[$i]\" \n";
              print STDERR "Gold Standard was: \"$gsData[$i]\" \n";
              $exitCode=2;
              last;
            }
            if ( ( abs($testData[1]) <= $zeroTol ) && ( abs($goldData[1]) <= $zeroTol ) )
	    {
              # no op since both test and gold data are less than the zero tolerance
	    }
            else
            {
	      $absDiff = abs($testData[1] - $goldData[1]);
              $relDiff = $absDiff / abs($goldData[1]);   
              if ( ( $absDiff < $absTol ) && ( $relDiff < $relTol ) )
              {
                # two numeric fields match to within specified absolute tolerance and relative tolerance
              }
              else
              {
                print STDERR "Numeric comparison failed in column $i on line $lineCount: \n";
                print STDERR "File $testFileName had \"$testFileData[$i]\" while Gold Standard had \"$gsData[$i]\" \n";
                print STDERR "Calculated absDiff and relDiff = ($absDiff,$relDiff)\n";
                $exitCode=2;
                last;
              }
            }
          }
          else
          {
            # this handles date elements from lines that don't start with XBEGIN.  
            # There are three cases, as noted in the comments below.
	    if ( ($foundDataStart > 0) && ($testFileData[$i] =~ /:/ ) && ($gsData[$i] =~/:/) )
            { 
              # these are fields that contain time points.  They are of the form  0.0001:1
              # where 0.0001 is the value and 1 is the idx.  This next block breaks them
              # apart into two pieces, each of which is tested separately.  These values are
              # found after the #C line
              @testData = split(/:/,$testFileData[$i]);
              @goldData = split(/:/,$gsData[$i]); 
              foreach $idx (1 ..2)
              {
                if ( ( abs($testData[$idx]) <= $zeroTol ) && ( abs($goldData[$idx]) <= $zeroTol ) )
	        {
                  # no op since both test and gold data are less than the zero tolerance
	        }
                else
                {
	          $absDiff = abs($testData[$idx] - $goldData[$idx]);
                  $relDiff = $absDiff / abs($goldData[$idx]);   
                  if ( ( $absDiff < $absTol ) && ( $relDiff < $relTol ) )
                  {
                    # two numeric fields match to within specified absolute tolerance and relative tolerance
                  }
                  else
                  {
                    print STDERR "Numeric comparison failed in column $i on line $lineCount: \n";
                    print STDERR "File $testFileName had \"$testFileData[$i]\" while Gold Standard had \"$gsData[$i]\" \n";
                    print STDERR "Calculated absDiff and relDiff = ($absDiff,$relDiff)\n"; 
                    $exitCode=2;
                    last;
                  }
		}
                if ($exitCode != 0) { last;}
              } 
            }
            elsif ( looks_like_number($testFileData[$i]) && looks_like_number($gsData[$i]) )
            {  
              # this block is for numbers that didn't match the value:idx format above
              if ( ( abs($testFileData[$i]) <= $zeroTol ) && ( abs($gsData[$i]) <= $zeroTol ) )
	      {
                # no op since both test and gold data are less than the zero tolerance
	      }
              else
              {
	        $absDiff = abs($testFileData[$i] - $gsData[$i]);
                $relDiff = $absDiff / abs($gsData[$i]);   
                if ( ( $absDiff < $absTol ) && ( $relDiff < $relTol ) )
                {
                  # two numeric fields match to within specified absolute tolerance and relative tolerance
                  # debug info, to verify that the comparison is working okay
                  #print STDERR "Numeric comparison for column $i on line $lineCount: \n";
                  #print STDERR "File $testFileName had \"$testFileData[$i]\" while Gold Standard had \"$gsData[$i]\" \n";
                  #print STDERR "Calculated absDiff and relDiff = ($absDiff,$relDiff)\n"; 
                }
                else
                {
                  print STDERR "Numeric comparison failed in column $i on line $lineCount: \n";
                  print STDERR "File $testFileName had \"$testFileData[$i]\" while Gold Standard had \"$gsData[$i]\" \n";
                  print STDERR "Calculated absDiff and relDiff = ($absDiff,$relDiff)\n"; 
                  $exitCode=2;
                  last;
                }
              }
            }
            elsif( ($testFileData[$i] eq $gsData[$i]) ) 
            {
              # two string fields match
              # debug info to verify that the string comparison is working okay
              #print STDERR "$testFileData[$i] same as $gsData[$i]\n";
            }
            else
            {
              print STDERR "String elements failed compare on line $lineCount:\n";
              print STDERR "File $testFileName produced: \"$testFileData[$i]\" \n";
              print STDERR "Gold Standard was: \"$gsData[$i]\" \n";
              $exitCode=2;
              last;
            }
          }                 
        }     
      }
    }

    # close files
    close(GSFILE);
    close(TESTFILE);
  }

  # check the return values
  if ( $exitCode != 0 )
  {
    print "test Failed!\n";
    print "Exit code = $exitCode\n";
    exit $exitCode;
  }
  else
  {
    print "test passed for $idx variables on .PRINT line\n";
  }
}

# final check on the return values
if ( $exitCode != 0 )
{
  print "test Failed!\n";
  print "Exit code = $exitCode\n";
  exit $exitCode;
}
else
{
  print "test passed!\n";
  print "Exit code = $exitCode\n";
  exit $exitCode;
}
