#!/usr/bin/env perl

use lib '..';
use XyceRegression::Tools;
use MeasureCommon;

$Tools = XyceRegression::Tools->new();
#$Tools->setDebug(1);
#$Tools->setVerbose(1);

# The input arguments to this script are: 
# $1 = location of Xyce binary
# $2 = location of xyce_verify.pl script
# $3 = location of compare script 
# $4 = location of circuit file to test
# $5 = location of gold standard prn file

# If Xyce does not produce a prn file, then we return exit code 10.
# If Xyce succeeds, but the test fails, then we return exit code 2.
# If the shell script fails for some reason, then we return exit code 1.
# Otherwise we return the exit code of compare or xyce_verify.pl

# Since the shell script runs Xyce and the comparison program, it is
# responsible for capturing any error output from Xyce and the STDOUT & STDERR
# from the comparison program.  The outside script, run_xyce_regression,
# expects the STDERR output from Xyce to go into $CIRFILE.err, the STDOUT
# output from comparison to go into $CIRFILE.prn.out and the STDERR output from
# comparison to go into $CIRFILE.prn.err.  

use Getopt::Long;
use Scalar::Util qw(looks_like_number);

&GetOptions( "verbose!" => \$verbose );
$XYCE=$ARGV[0];
$XYCE_VERIFY=$ARGV[1];
#$XYCE_COMPARE=$ARGV[2];
$CIRFILE=$ARGV[3];
$GOLDPRN=$ARGV[4];

# these are the tolerances used in comparing the measures calculated by this 
# script to those calculated by Xyce.  Zerotol is needed because some measured
# values should be zero (and reltol won't make sense in that case).

# phase is output in degrees and for fourier components with very small magnitude, 
# it can be have several degrees of scatter.  Thus it gets its own abstol
my $phaseAbsTol = 1.0;
my $relTol = 0.01;
my $zeroTol = 1.0e-8;
my $expectedDoubleCount=51;
my $defaultPrecision=6;
my $numSteps=3;

#
# this subroutine is used to compare gold standard circuit.cir.mt? files 
# to what the simulator creates
sub compareFourierMeasureFiles
{
  my ($goldFile, $simFile) = @_;
  #
  # Now look for the measure output file and compare it to a 
  # gold standard line by line.
  #
  open(RESULTS, $simFile);
  open(GOLD_STD, $goldFile);
  my $lineNumber=1;
  my $retval = 0;
  my $doubleCount=0;
  my $line;
  my $line_gs;
  while( ($line=<RESULTS>) && ($line_gs=<GOLD_STD>) )
  {
    # process a line into text and values.
    chop $line;
    # Remove leading spaces on line, otherwise the spaces become 
    # element 0 of "@lineOfDataFromXyce" instead of the first column of data.
    $line =~ s/^\s*//;
    my @lineOfDataFromXyce = (split(/[\s,]+/, $line));
    #print "line of data from Xyce = @lineOfDataFromXyce\n";
    #print "end count = $#lineOfDataFromXyce\n";
    
    # process a line_gs into text and values.
    chop $line_gs;
    $line_gs =~ s/^\s*//;
    my @gsLineOfDataFromXyce = (split(/[\s,]+/, $line_gs));
    
    if( $#lineOfDataFromXyce != $#gsLineOfDataFromXyce )
    {
      print "Xyce's output to measure file doesn't match the gold standard at line $lineNumber\n";
      print "Xyce's output: $line\n";
      print "Gold Standard: $line_gs\n";
      $retval=2;
    }
    else
    {
      # the two files have the same number of items on a line.  
      # compare individual values as scalars  This will have the 
      # effect fo 
      for( $i=0; $i<=$#lineOfDataFromXyce; $i++ )
      {
        #print "data item = $lineOfDataFromXyce[$i]\n";
        if (looks_like_number($lineOfDataFromXyce[$i]) && looks_like_number($gsLineOfDataFromXyce[$i] ) )
        {
          if( ( abs( $lineOfDataFromXyce[$i] ) < $zeroTol) &&  (abs( $gsLineOfDataFromXyce[$i] ) < $zeroTol ))
          {
             #print "number compare below zeroTol $lineOfDataFromXyce[$i] , $gsLineOfDataFromXyce[$i] ok\n";
          }
          elsif( (($i==3) || ($i==5)) && (abs( $lineOfDataFromXyce[$i] - $gsLineOfDataFromXyce[$i] ) < $phaseAbsTol )) 
          {
             # phase needs different handling
             #print "Phase comparison passed $lineOfDataFromXyce[$i] , $gsLineOfDataFromXyce[$i]\n";
          }
          elsif( ( (abs( $lineOfDataFromXyce[$i] - $gsLineOfDataFromXyce[$i] ))/abs($gsLineOfDataFromXyce[$i]) < $relTol )) 
          {
             # regular compare
          }
          else
          {
            print "On line $lineNumber of $CIRFILE.mt0 and $GOLDPRN found a numeric difference\n";
            print "Comparing Xyce's \"$lineOfDataFromXyce[$i]\" to GS \"$gsLineOfDataFromXyce[$i]\" as ";
            print "Failed numeric compare\n";
            $retval=2;
            last;
          }
          if( $lineOfDataFromXyce[$i] =~ /e/)
          {
	    $doubleCount++;
            #print "double count and data item = $doubleCount and $lineOfDataFromXyce[$i]\n";
            $retval = MeasureCommon::checkNumberFormat($lineOfDataFromXyce[$i],1,$defaultPrecision);
            if ( $retval != 0 )
            {
            print "test Failed because of number format!\n";
            $retval=2;
            last;
          }
        }
        }
        elsif( $lineOfDataFromXyce[$i] eq $gsLineOfDataFromXyce[$i] )
        {
          # same in string context so ok.
          # print "string compare ok\n";
        }
        else
        {
          print "Elements failed compare on line $lineNumber: \n";
          print "Xyce produced: \"$lineOfDataFromXyce[$i]\"\n";
          print "Gold standard: \"$gsLineOfDataFromXyce[$i]\"\n";
          $retval=2;
          last;
        }
      
      }
    }
    $lineNumber++;
  }
  close(RESULTS);
  close(GOLD_STD);

  if ( $doubleCount != $expectedDoubleCount )
  {
    print "Found $doubleCount doubles.  Expected $expectedDoubleCount doubles.\n";
    $retval=2; 
  }
  return $retval;
}


# change ending on gold standard file from ".prn" as passed in to ".mt0" 
# which is the file storing the measure results.

$GOLDPRN =~ s/.prn$//;

if (defined($verbose)) { $Tools->setVerbose(1); }

sub verbosePrint { $Tools->verbosePrint(@_); }

$retval = -1;
$retval=$Tools->wrapXyce($XYCE,$CIRFILE);
if ($retval != 0) { print "Exit code = $retval\n"; exit $retval; }
if (not -s "$CIRFILE.prn" ) { print "Exit code = 14\n"; exit 14; }

#
# Did we make a measure file
#
if (not -s "$CIRFILE.mt0" ) { print "Exit code = 17\n"; exit 17; }

#
# Now we need to count the number of .MEASURE statemets in the netlist
# and make sure we have some output for each one in the output file
#
open(NETLIST, "$CIRFILE");
$numMeasures=0;
# these are used to store parameters which can limit or windows the scope of a measure
my @AtVals = (0.0, 0.0);
my @NumFreqVals = (0.0, 0.0);
my @GridSizeVals = (0.0, 0.0);
my @measureNames = ("", "");
while( $line=<NETLIST> )
{
  if( $line =~ /^\.measure/i)
  {  
    $numMeasures++;
    # try to parse out ON= and OFF= values 
    @words = split( /\s+/, $line);
    @measureNames[ $numMeasures-1 ] = $words[2];
    foreach $word (@words)
    {
      if( $word =~ "=" )
      {
        ($keyword, $value) = split( /=/, $word);
        if( $keyword eq "AT" )
        {
          @AtVals[$numMeasures-1] = $value;
        }
        elsif( $keyword eq "NUMFREQ" )
        {
          @NumFreqVals[$numMeasures-1] = $value;
        }
        elsif( $keyword eq "GRIDSIZE" )
        {
          @GridSizeVals[$numMeasures-1] = $value;
        }
        elsif( $keyword eq "TO" )
        {
          @ToVals[$numMeasures-1] = $value;
        }
        elsif( $keyword eq "MINVAL" )
        {
          @minVals[$numMeasures-1] = $value;
        }
        elsif( $keyword eq "RISE" )
        {
          @riseVals[$numMeasures-1] = $value;
        }
        else
        {
          # not great, but use to catch caes of 
          # when v(1)=some_value
          @findVals[$numMeasures-1] = $value;
        }
      }
    }
  }
}
close(NETLIST);

#
# Make sure the measurement was not output multiple times in parallel
# in either the output file or the measure file.
$numMeasOut = `grep -ic $measureNames[ $i ] $CIRFILE.out`;
if ( $numMeasOut != $numSteps )
{
   print "Xyce printed $measureNames[ $i ] measure to standard output incorrect number of times.\n";
   print "Number of Steps and Number of Measure Output = $numSteps, $numMeasOut\n";
   $retval = 2;
}
$numMeasMT0 = `grep -ic $measureNames[ $i ] $CIRFILE.mt0`;
$numMeasMT1 = `grep -ic $measureNames[ $i ] $CIRFILE.mt1`;
$numMeasMT2 = `grep -ic $measureNames[ $i ] $CIRFILE.mt2`;
if ( ($numMeasMT0 != 1) || ($numMeasMT1 != 1) || ($numMeasMT0 != 1) )
{
  print "Xyce printed out $measureNames[ $i ] measure to file multiple times.\n";
  $retval = 2;
  verbosePrint "test Failed!\n";
  print "Exit code = $retval\n"; 
  exit $retval;
}

# now compare the gold standards and the measure file contents
$retval = compareFourierMeasureFiles("$GOLDPRN.mt0", "$CIRFILE.mt0");
if ( $retval != 0 )
{
  verbosePrint "test Failed!\n";
  print "Exit code = $retval\n"; 
  exit $retval;
}

$retval = compareFourierMeasureFiles("$GOLDPRN.mt1", "$CIRFILE.mt1");
if ( $retval != 0 )
{
  verbosePrint "test Failed!\n";
  print "Exit code = $retval\n"; 
  exit $retval;
}
  
$retval = compareFourierMeasureFiles("$GOLDPRN.mt2", "$CIRFILE.mt2");
if ( $retval != 0 )
{
  verbosePrint "test Failed!\n";
  print "Exit code = $retval\n"; 
  exit $retval;
}
  
verbosePrint "test Passed!\n";
print "Exit code = $retval\n"; 
exit $retval;


